/** 
 *   Camera Control for STAR I (originally).
 *
 *   Included in libiarcontrol v 0.1.5
 *
 *   (c) Copyright 2005 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 **/
#define _GNU_SOURCE

#include <string.h>
#include <stdlib.h>
#include <wordexp.h>
#include "my_stdio.h"
#include "parse_tools.h"
#include "locks.h"


static struct configuration *make_conf(void)
{
	struct configuration *fc;

	fc = (struct configuration *) malloc(sizeof(struct configuration));
	if (fc == NULL) {
		SYSLOG(LOG_ERR,
		       "struct configuration *make_conf: malloc fail.");
		/* FIXME: */
		return NULL;
	}
	fc->flag = FCF_INITIAL;
	fc->type = FCT_EOF;
	fc->next = NULL;
	fc->pte_data = NULL;
	fc->expand = NULL;
	memset(fc->line, 0, LINE_LEN_CONF);
	
	return fc;	
}


int read_conf_file(struct configuration **conf, char *filename)
{
	struct configuration *fc;
	FILE *fp;
	int count = 0;

	fp = fopen(filename, "r");
	if( fp == NULL ){
                SYSLOG(LOG_ERR,
                       "read_conf_file: Error opening file: %s",
		       filename);
                return -1;
        }

	if (*conf == NULL) {
		fc = *conf = make_conf();
	} else {
		for(fc = *conf; fc->next != NULL; fc = fc->next);
	}

	while( fgets(fc->line, LINE_LEN_CONF, fp)){
		count++;
		fc->line_num = count;
		fc->len = strlen(fc->line);
		if(fc->line[fc->len -1] == '\n') {
			fc->line[fc->len -1] = 0;
			while (fc->line[fc->len -2] == '\\' &&
			       fgets(fc->line+fc->len-2, 
				     LINE_LEN_CONF - fc->len, fp) ){
				fc->len = strlen(fc->line);
			}
				
		}else{	/* Line to long */
			SYSLOG(LOG_WARNING, "%s: line %d possible to long.", 
			       filename, count);
		}
		fc->sline = stripwhite(fc->line);
		if (parse_line(fc->sline, &fc->tag, 
			       &fc->data, &fc->comment) == 0){
			fc->type = FCT_TAG;
		} else if(fc->comment != NULL) {
			fc->type = FCT_COMMENT;
		} else if (*fc->sline == 0) {
			fc->type = FCT_NULL;
		} else {
			fc->type = FCT_UNKNOW;
			SYSLOG(LOG_ERR,"%s: ignoring line %d \"%s\"",
			      filename, fc->line_num, fc->line);
		}
		fc->next = make_conf();
		fc = fc->next;
	}

	fclose(fp);

	return 0;
}


static void __free_conf(struct configuration *conf)
{
	if (conf->flag & FCF_ENVIRON)
		putenv (conf->tag);
	if (conf->flag & FCF_EXPAND)
		free (conf->expand);

	free (conf);
}


int free_conf(struct configuration *conf)
{
	struct configuration *cf = conf;
	struct configuration *cfp;

	while(cf != NULL){
		cfp = cf;
		cf = cf->next;
		__free_conf(cfp);
	}

	return 0;
}


int write_conf_file(struct configuration *conf, char *filename)
{
	struct configuration *cf = conf;
	FILE *fp = NULL;
	char fmt[40];
	int c, i, t;
	int lock;

	if ( (lock = if_lock_file(filename)) ) {
		SYSLOG(LOG_ERR,"File %s locked by pid: %d.", filename, lock);
		return -1;
	}

	//fp = stdout;
	fp = fopen(filename, "w");
	if( fp == NULL ){
                SYSLOG(LOG_ERR,
                       "write_conf_file: Error opening file: %s",
		       filename);
                return -1;
        }

	while (cf != NULL) {
		if (!(cf->flag &FCF_DELETED)) {
		if (cf->type == FCT_TAG) {
			if ( cf->comment != NULL ) {
				c = sprintf(fmt,"%%s = %%s");
				i = 40-strlen(cf->tag)-strlen(cf->data)-3;
				t = i >> 3;
				if ( i & 0x07 ) t++;
				for(i=0; i < t; i++)
					c += sprintf(fmt+c,"\t");
				sprintf(fmt+c,"# %%s\n");
				fprintf(fp, fmt,
					cf->tag, cf->data, cf->comment);
			} else {
				fprintf(fp,"%s = %s\n", cf->tag, cf->data);
			}
		} else if (cf->type == FCT_COMMENT) {
			fprintf(fp,"#%s\n", cf->comment);
		} else if (cf->type == FCT_NULL) {
			fprintf(fp,"\n");
		} else if (!(cf->flag & FCF_DELETED) && cf->type != FCT_EOF) {
			fprintf(fp,"# [ignored: %s]\n", cf->sline);
		}
		}
		cf = cf->next;
	}
	fclose(fp);
	unlock_file(filename);

	return 0;
}


struct configuration *find_conf_tag(struct configuration *conf, char *tag)
{
	struct configuration *cfp = conf;

	while(cfp != NULL) {
		if (cfp->type == FCT_TAG && !(cfp->flag & FCF_DELETED)) {
			if (strcmp(cfp->tag, tag) == 0 ) {
				return cfp;
			}
		}
		cfp = cfp->next;
	}

	return (struct configuration *) NULL;
}


static int write_conf_line(char *line, char *tag, char *data, char *comment)
{
	int len;
	
	if ( comment != NULL ) {
		len = sprintf(line, "%s = %s # %s", tag, data, comment);
	} else {
		len = sprintf(line,"%s = %s", tag, data);
	}

	return len;
}


int mod_conf_data(struct configuration *cf, char *data)
{
	int ret = -1;
	char line[LINE_LEN_CONF];
	
	if (cf != NULL && cf->type == FCT_TAG && !(cf->flag & FCF_DELETED)) {
		//memset(line, 0, LINE_LEN_CONF);
		cf->len = write_conf_line(line, cf->tag, data, cf->comment);
		if (cf->len > LINE_LEN_CONF) cf->len = LINE_LEN_CONF;
		memset(cf->line, 0, LINE_LEN_CONF);
		strncpy(cf->line, line, cf->len);
		cf->sline = cf->line;
		if (parse_line(cf->line, &cf->tag, 
			       &cf->data, &cf->comment) == 0){
			ret = 0;
		}
	}

	return ret;
}


int change_conf_tag(struct configuration *conf, char *tag, char *data)
{
	int ret = -1;
	struct configuration *t = find_conf_tag(conf, tag);
	
	if (t != NULL) {
		ret = mod_conf_data(t, data);
	}

	return ret;
}


int add_conf_tag(struct configuration *conf, 
			  char *tag, char *data, char *comment)
{
	struct configuration *p = conf;

	while(p->next != NULL) p = p->next;
	if ( p->type != FCT_EOF )
		return -1;

	p->next = make_conf();
	p->type = FCT_TAG;
	p->len = write_conf_line(p->line, tag, data, comment);
	p->sline = p->line;
	parse_line(p->line, &p->tag, &p->data, &p->comment);

	return 0;
}


/* Only mark which FCF_DELETED, don't delete this */
int del_conf_tag(struct configuration *conf, char *tag)
{
	int ret = -1;
	struct configuration *t = find_conf_tag(conf, tag);
	
	if (t != NULL) {
		t->flag |= FCF_DELETED;
		ret = 0;
	}

	return ret;
}


int purge_conf_tag(struct configuration *conf)
{
	struct configuration *cf = conf;
	struct configuration *cfp = NULL;
	struct configuration *cfd;

	while (conf->flag & FCF_DELETED) { /* los primeros */
		cfd = conf;
		conf = conf->next;
		__free_conf(cfd);
	}
	if(conf != NULL) {
		cfp = conf;
		cf = conf->next;
		while(cf != NULL){ /* el el medio */
			if (cf->flag & FCF_DELETED) {
				cfd = cf;
				cfp->next = cf->next;
				cf = cf->next;
				__free_conf(cfd);
			} else {
				cfp = cf;
				cf = cf->next;
			}
		}
	}

	return 0;
}


/* Adds definitions from the environment, and expand it. */

int put_env_conf(struct configuration *conf)
{
	struct configuration *cfp = conf;
	wordexp_t result;
	int i, ret;
	int len;

	while(cfp != NULL) {
		if (cfp->type == FCT_TAG && cfp->expand == NULL  &&
		    !(cfp->flag & FCF_DELETED)) {
			
			if(cfp->flag & FCF_ENVIRON) {
				cfp->flag &= ~FCF_ENVIRON;
				putenv (cfp->tag);
			}
			if(cfp->flag & FCF_EXPAND) {
				cfp->flag &= ~FCF_EXPAND;
				free(cfp->expand);
			}
			len = strlen(cfp->tag) + 1;
			if( (ret = wordexp (cfp->data, &result, 0)) == 0) {
				for(i=0;i<result.we_wordc;i++)
					len += strlen(result.we_wordv[i])+1;
				/* len tiene 1 mas por el \0 del
                                   sprintf que sigue */
				cfp->expand = (char *) 
					malloc(sizeof(char) * len);
				cfp->flag |= FCF_EXPAND;

				len = sprintf(cfp->expand,"%s=", cfp->tag);
				cfp->pte_data = cfp->expand+len;
				len += sprintf(cfp->expand+len, "%s", 
					       result.we_wordv[0]);
				for(i=1;i<result.we_wordc;i++){
					len += sprintf(cfp->expand+len, " %s", 
						       result.we_wordv[i]);
				}
				wordfree (&result);
				if (putenv(cfp->expand) == 0) {
					cfp->flag |= FCF_ENVIRON;
				}
			}else if (ret == WRDE_NOSPACE ) {
				wordfree (&result);
			}else {
				SYSLOG(LOG_ERR,"Error expanding words: %s",
				       cfp->data);

			}
		}
		cfp = cfp->next;
	}

	return 0;
}


int unset_env_conf(struct configuration *conf)
{
	struct configuration *cfp = conf;

	while(cfp != NULL) {
		if (cfp->flag & FCF_ENVIRON) {
			cfp->flag &= ~FCF_ENVIRON;
			putenv (cfp->tag);
		}
		cfp = cfp->next;
	}

	return 0;


}
